package com.konaworks.idea.inspections.javaee.ejb;

import com.intellij.codeInspection.InspectionManager;
import com.intellij.codeInspection.LocalQuickFix;
import com.intellij.codeInspection.ProblemDescriptor;
import com.intellij.codeInspection.ProblemHighlightType;
import com.intellij.javaee.model.xml.DisplayName;
import com.intellij.javaee.model.xml.SecurityRole;
import com.intellij.javaee.model.xml.ejb.EjbJar;
import com.intellij.javaee.model.xml.ejb.EjbRelation;
import com.intellij.javaee.model.xml.ejb.EnterpriseBeans;
import com.intellij.javaee.model.xml.ejb.EntityBean;
import com.intellij.javaee.model.xml.ejb.MessageDrivenBean;
import com.intellij.javaee.model.xml.ejb.MethodPermission;
import com.intellij.javaee.model.xml.ejb.SessionBean;
import com.intellij.openapi.project.Project;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.XmlElement;
import com.intellij.psi.xml.XmlText;
import com.intellij.util.IncorrectOperationException;
import com.intellij.util.xml.DomFileElement;
import com.intellij.util.xml.GenericDomValue;
import com.konaworks.idea.inspections.javaee.Bundle;
import com.konaworks.idea.inspections.javaee.DuplicateNamedElementsProcessor;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.util.ArrayList;
import java.util.List;

public class EjbJarDuplicatesInspection extends AbstractEjbJarInspection {
  private JPanel optionsPanel;
  private JCheckBox ejbNameCheckBox;
  private JCheckBox ejbDisplayNameCheckBox;
  private JCheckBox relationNameCheckBox;
  private JCheckBox methodPermissionSecurityRoleCheckBox;

  // settings readible and writable DefaultJDOMExternalizer
  public boolean checkEjbNames = true;
  public boolean checkEjbDisplayNames = true;
  public boolean checkRelationNames = true;
  public boolean checkMethodPermissionSecurityRoles = true;

  public EjbJarDuplicatesInspection() {
    ejbNameCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkEjbNames = ejbNameCheckBox.isSelected();
      }
    });
    ejbDisplayNameCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkEjbDisplayNames = ejbDisplayNameCheckBox.isSelected();
      }
    });
    relationNameCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkRelationNames = relationNameCheckBox.isSelected();
      }
    });
    methodPermissionSecurityRoleCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkMethodPermissionSecurityRoles = methodPermissionSecurityRoleCheckBox.isSelected();
      }
    });
  }

  @NotNull
  public String getDisplayName() {
    return Bundle.message("inspection.name.KW_EjbJarDuplicatesInspection");
  }

  @NotNull
  public String getShortName() {
    return "KW_EjbJarDuplicatesInspection";
  }


  @Nullable
  public JComponent createOptionsPanel() {
    ejbNameCheckBox.setSelected(checkEjbNames);
    ejbDisplayNameCheckBox.setSelected(checkEjbDisplayNames);
    relationNameCheckBox.setSelected(checkRelationNames);
    methodPermissionSecurityRoleCheckBox.setSelected(checkMethodPermissionSecurityRoles);
    
    return optionsPanel;
  }

  protected ProblemDescriptor[] checkDomFile(@NotNull DomFileElement<EjbJar> domFileElement, @NotNull InspectionManager manager, boolean isOnTheFly) {
    final List<ProblemDescriptor> problems = new ArrayList<ProblemDescriptor>();
    final EjbJar ejbJar = domFileElement.getRootElement();

    if (checkEjbNames || checkEjbDisplayNames) {
      final DuplicateNamedElementsProcessor<GenericDomValue<String>> ejbNameProcessor = new DuplicateNamedElementsProcessor<GenericDomValue<String>>();
      final DuplicateNamedElementsProcessor<DisplayName> displayNameProcessor = new DuplicateNamedElementsProcessor<DisplayName>();
      // todo for the display names: see the remark in the WebAppWarningDuplicatesInspection class

      final EnterpriseBeans enterpriseBeans = ejbJar.getEnterpriseBeans();
      for (EntityBean entityBean : enterpriseBeans.getEntities()) {
        if (checkEjbNames) {
          final GenericDomValue<String> domEjbName = entityBean.getEjbName();
          final String ejbName = domEjbName.getStringValue();
          if (ejbName != null) {
            ejbNameProcessor.put(ejbName, domEjbName);
          }
        }
        if (checkEjbDisplayNames) {
          for (DisplayName domDisplayName : entityBean.getDisplayNames()) {
            displayNameProcessor.put(domDisplayName.getValue(), domDisplayName);
          }
        }
      }

      for (MessageDrivenBean messageDrivenBean : enterpriseBeans.getMessageDrivens()) {
        if (checkEjbNames) {
          final GenericDomValue<String> domEjbName = messageDrivenBean.getEjbName();
          final String ejbName = domEjbName.getStringValue();
          if (ejbName != null) {
            ejbNameProcessor.put(ejbName, domEjbName);
          }
        }
        if (checkEjbDisplayNames) {
          for (DisplayName domDisplayName : messageDrivenBean.getDisplayNames()) {
            displayNameProcessor.put(domDisplayName.getValue(), domDisplayName);
          }
        }
      }

      for (SessionBean sessionBean : enterpriseBeans.getSessions()) {
        if (checkEjbNames) {
          final GenericDomValue<String> domEjbName = sessionBean.getEjbName();
          final String ejbName = domEjbName.getStringValue();
          if (ejbName != null) {
            ejbNameProcessor.put(ejbName, domEjbName);
          }
        }
        if (checkEjbDisplayNames) {
          for (DisplayName domDisplayName : sessionBean.getDisplayNames()) {
            displayNameProcessor.put(domDisplayName.getValue(), domDisplayName);
          }
        }
      }
      // end of duplicates retrieval...

      if (checkEjbNames) {
        for (GenericDomValue<String> duplicate : ejbNameProcessor.getDuplicates()) {
          final XmlElement xmlElement = duplicate.getXmlElement();
          if (xmlElement != null) {
            final XmlText xmlText = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
            if (xmlText != null) {
              problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("duplicate.ejb-name"), REMOVE_QUICK_FIX, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
            }
          }
        }
      }
      if (checkEjbDisplayNames) {
        for (DisplayName duplicate : displayNameProcessor.getDuplicates()) {
          final XmlElement xmlElement = duplicate.getXmlElement();
          if (xmlElement != null) {
            final XmlText xmlText = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
            if (xmlText != null) {
              problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("duplicate.display-name"), REMOVE_QUICK_FIX, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
            }
          }
        }
      }
    }

    if (checkRelationNames) {
      final DuplicateNamedElementsProcessor<GenericDomValue<String>> processor = new DuplicateNamedElementsProcessor<GenericDomValue<String>>();

      for (EjbRelation ejbRelation : ejbJar.getRelationships().getEjbRelations()) {
        final GenericDomValue<String> domEjbRelName = ejbRelation.getEjbRelationName();
        final String ejbRelationName = domEjbRelName.getStringValue();
        if (ejbRelationName != null) {
          processor.put(ejbRelationName, domEjbRelName);
        }
      }

      for (GenericDomValue<String> duplicate : processor.getDuplicates()) {
        final XmlElement xmlElement = duplicate.getXmlElement();
        if (xmlElement != null) {
          final XmlText xmlText = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
          if (xmlText != null) {
            problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("duplicate.ejb-relation-name"), REMOVE_QUICK_FIX, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
          }
        }
      }
    }

    if (checkMethodPermissionSecurityRoles) {
      final List<MethodPermission> methodPermissions = ejbJar.getAssemblyDescriptor().getMethodPermissions();
      for (MethodPermission methodPermission : methodPermissions) {
        final DuplicateNamedElementsProcessor<GenericDomValue<SecurityRole>> processor = new DuplicateNamedElementsProcessor<GenericDomValue<SecurityRole>>();

        final List<GenericDomValue<SecurityRole>> roles = methodPermission.getRoleNames();
        for (GenericDomValue<SecurityRole> securityRole : roles) {
          processor.put(securityRole.getStringValue(), securityRole);
        }

        for (GenericDomValue<SecurityRole> duplicate : processor.getDuplicates()) {
          final XmlElement xmlElement = duplicate.getXmlElement();
          if (xmlElement != null) {
            final XmlText xmlText = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
            if (xmlText != null) {
              problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("duplicate.role-name"), REMOVE_QUICK_FIX, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
            }
          }
        }
      }
    }

    return problems.toArray(new ProblemDescriptor[problems.size()]);
  }

  private LocalQuickFix REMOVE_QUICK_FIX = new RemoveQuickFix();

  private static class RemoveQuickFix implements LocalQuickFix {
    @NotNull
    public String getName() {
      return Bundle.message("quickfix.remove_duplicate_role-name");
    }

    public void applyFix(@NotNull Project project, ProblemDescriptor descriptor) {
      try {
        descriptor.getPsiElement()/*XmlText*/.getParent().delete();
      } catch (IncorrectOperationException e) {
        // no actions
      }
    }

    @NotNull
    public String getFamilyName() {
      return getName();
    }
  }
}
