package com.konaworks.idea.inspections.javaee.ejb;

import com.intellij.codeInspection.InspectionManager;
import com.intellij.codeInspection.LocalQuickFix;
import com.intellij.codeInspection.ProblemDescriptor;
import com.intellij.codeInspection.ProblemHighlightType;
import com.intellij.javaee.model.xml.SecurityRole;
import com.intellij.javaee.model.xml.SecurityRoleRef;
import com.intellij.javaee.model.xml.ejb.EjbJar;
import com.intellij.javaee.model.xml.ejb.EntityBean;
import com.intellij.javaee.model.xml.ejb.MethodPermission;
import com.intellij.javaee.model.xml.ejb.SessionBean;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiFile;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.XmlElement;
import com.intellij.psi.xml.XmlFile;
import com.intellij.psi.xml.XmlTag;
import com.intellij.psi.xml.XmlText;
import com.intellij.util.IncorrectOperationException;
import com.intellij.util.xml.DomFileElement;
import com.intellij.util.xml.DomManager;
import com.intellij.util.xml.GenericDomValue;
import com.konaworks.idea.inspections.javaee.Bundle;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.List;

public class EjbJarSecurityRoleReferenceInspection extends AbstractEjbJarInspection {

  @NotNull
  public String getDisplayName() {
    return Bundle.message("inspection.name.KW_EjbJarSecurityRoleReferenceInspection");
  }

  @NotNull
  @NonNls
  public String getShortName() {
    return "KW_EjbJarSecurityRoleReferenceInspection";
  }

  protected ProblemDescriptor[] checkDomFile(@NotNull DomFileElement<EjbJar> domFileElement, @NotNull InspectionManager manager, boolean isOnTheFly) {
    List<ProblemDescriptor> problems = new ArrayList<ProblemDescriptor>();

    List<String> roleDefs = new ArrayList<String>();
    for (SecurityRole securityRole : domFileElement.getRootElement().getAssemblyDescriptor().getSecurityRoles()) {
      roleDefs.add(securityRole.getRoleName().getValue());
    }

    // 1 security-role-ref in session
    // note for 3.0 ejb-jar.xml files this security-role-ref check is already done by IDEA (through schema validation,
    // but doesn't work very smoothly btw. and it is only a twirly under ejb-jar element)
    // todo maybe exclude this one for ejb 3.0 xml files ? // if (Float.valueOf(ejbJar.getVersion().getValue()) < 3)
    for (SessionBean sessionBean : domFileElement.getRootElement().getEnterpriseBeans().getSessions()) {
      for (SecurityRoleRef securityRoleRef : sessionBean.getSecurityRoleRefs()) {
        if (securityRoleRef.getRoleLink() != null && !roleDefs.contains(securityRoleRef.getRoleLink().getValue())) {
          final XmlElement roleLinkXmlElement = securityRoleRef.getRoleLink().getXmlElement();

          if (roleLinkXmlElement != null) {
            final XmlText xmlText = PsiTreeUtil.getChildOfAnyType(roleLinkXmlElement, XmlText.class);

            if (xmlText != null) {
              problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("undefined.role-name"),
              new LocalQuickFix[]{ADD_QUICK_FIX}, ProblemHighlightType.LIKE_UNKNOWN_SYMBOL));
              // the remove quick fix is not supported here, because the use of it would always yield in an incorrect security-role-ref.
            }
          }
        }
      }
    }

    // 2 security-role-ref in entity
    // note for 3.0 ejb-jar.xml files this security-role-ref check is already done by IDEA (through schema validation,
    // but doesn't work very smoothly btw. and it is only a twirly under ejb-jar element)
    // todo maybe exclude this one for ejb 3.0 xml files ? // if (Float.valueOf(ejbJar.getVersion().getValue()) < 3)
    for (EntityBean entityBean : domFileElement.getRootElement().getEnterpriseBeans().getEntities()) {
      for (SecurityRoleRef securityRoleRef : entityBean.getSecurityRoleRefs()) {
        if (!roleDefs.contains(securityRoleRef.getRoleLink().getValue())) {
          final XmlElement roleLinkXmlElement = securityRoleRef.getRoleLink().getXmlElement();

          if (roleLinkXmlElement != null) {
            final XmlText xmlText = PsiTreeUtil.getChildOfAnyType(roleLinkXmlElement, XmlText.class);

            if (xmlText != null) {
              problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("undefined.role-name"),
              new LocalQuickFix[]{ADD_QUICK_FIX}, ProblemHighlightType.LIKE_UNKNOWN_SYMBOL));
              // the remove quick fix is not supported here, because the use of it would always yield in an incorrect security-role-ref.
            }
          }
        }
      }
    }

    // 3 role-name in method-permission
    for (MethodPermission methodPermission : domFileElement.getRootElement().getAssemblyDescriptor().getMethodPermissions()) {
      for (GenericDomValue<SecurityRole> genericSecurityRole : methodPermission.getRoleNames()) {
        final String roleName = genericSecurityRole.getStringValue();
        if (!roleDefs.contains(roleName)) {
          final XmlElement securityRoleXmlElement = genericSecurityRole.getXmlElement();
          if (securityRoleXmlElement != null) {
            final XmlText xmlText = PsiTreeUtil.getChildOfAnyType(securityRoleXmlElement, XmlText.class);
            if (xmlText != null) {
              if ("*".equals(roleName)) {
                problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("specially.reserved.role-name.not.allowed"),
                    new LocalQuickFix[]{REMOVE_QUICK_FIX, REPLACE_WITH_UNCHECKED_QUICKFIX}, ProblemHighlightType.LIKE_UNKNOWN_SYMBOL));
              } else {
                // todo ? remove (already covered by JB, although without quickfix)
                problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("undefined.role-name"),
                    new LocalQuickFix[]{ADD_QUICK_FIX, REMOVE_QUICK_FIX}, ProblemHighlightType.LIKE_UNKNOWN_SYMBOL));
              }
            }
          }
        }
      }
    }

    return problems.toArray(new ProblemDescriptor[problems.size()]);
  }

  private LocalQuickFix ADD_QUICK_FIX = new AddQuickFix();

  private static class AddQuickFix implements LocalQuickFix {
    @NotNull
    public String getName() {
      return Bundle.message("quickfix.add.undefined.security-role");
    }

    public void applyFix(@NotNull Project project, ProblemDescriptor descriptor) {
      final XmlText xmlText = (XmlText) descriptor.getPsiElement();
      if (xmlText.getValue() != null) {
        final PsiFile file = descriptor.getPsiElement().getContainingFile();
        final DomFileElement<EjbJar> domFileElement = DomManager.getDomManager(project).getFileElement((XmlFile) file, EjbJar.class);
        if (domFileElement != null) {
          final EjbJar ejbJar = domFileElement.getRootElement();
          ejbJar.getAssemblyDescriptor().addSecurityRole().getRoleName().setValue(xmlText.getText());
        }
      }
    }

    @NotNull
    public String getFamilyName() {
      return getName();
    }
  }

  private LocalQuickFix REMOVE_QUICK_FIX = new RemoveQuickFix();

  private static class RemoveQuickFix implements LocalQuickFix {
    @NotNull
    public String getName() {
      return Bundle.message("quickfix.remove.undefined.security-role-reference");
    }

    public void applyFix(@NotNull Project project, ProblemDescriptor descriptor) {
      try {
        descriptor.getPsiElement()/*role-name XmlText*/.getParent()/*role-name XmlTag*/.delete();
      } catch (IncorrectOperationException e) {
        // no actions
      }
    }

    @NotNull
    public String getFamilyName() {
      return getName();
    }
  }

  private static final LocalQuickFix REPLACE_WITH_UNCHECKED_QUICKFIX = new ReplaceWithUncheckedQuickFix();

  private static class ReplaceWithUncheckedQuickFix implements LocalQuickFix {

    @NotNull
    public String getName() {
      return Bundle.message("quickfix.replace.role-name.elements.with.unchecked");
    }

    @NotNull
    public String getFamilyName() {
      return getName();
    }

    public void applyFix(@NotNull Project project, ProblemDescriptor descriptor) {
      // descriptor.getPsiElement is the role-name xmlElement
      final XmlTag methodPermissionXmlElement = (XmlTag) descriptor.getPsiElement()/*role-name XmlText*/.getParent()/*=role-name XmlTag*/.getParent();
      final MethodPermission methodPermission = (MethodPermission) DomManager.getDomManager(project).getDomElement(methodPermissionXmlElement);
      assert methodPermission != null;

      // delete role-name elems
      final List<GenericDomValue<SecurityRole>> roleNames = methodPermission.getRoleNames();
      try {
        // not using deleteChildRange here, to prevent cases where a silly user has mixed the role-name elements with other...
        for (GenericDomValue<SecurityRole> roleName : roleNames) {
          final XmlElement xmlElement = roleName.getXmlElement();
          if (xmlElement != null) {
            xmlElement.delete();
          }
        }
      } catch (IncorrectOperationException e) {
        // no actions
      }

      // 'add' the unchecked item
      methodPermission.getUnchecked().setValue(true);
    }
  }
}
