package com.konaworks.idea.inspections.javaee.ejb;

import com.intellij.codeInspection.InspectionManager;
import com.intellij.codeInspection.LocalQuickFix;
import com.intellij.codeInspection.ProblemDescriptor;
import com.intellij.codeInspection.ProblemHighlightType;
import com.intellij.javaee.model.xml.SecurityRole;
import com.intellij.javaee.model.xml.SecurityRoleRef;
import com.intellij.javaee.model.xml.ejb.EjbJar;
import com.intellij.javaee.model.xml.ejb.EntityBean;
import com.intellij.javaee.model.xml.ejb.MethodPermission;
import com.intellij.javaee.model.xml.ejb.SessionBean;
import com.intellij.openapi.project.Project;
import com.intellij.psi.xml.XmlElement;
import com.intellij.util.IncorrectOperationException;
import com.intellij.util.xml.DomFileElement;
import com.intellij.util.xml.GenericDomValue;
import com.konaworks.idea.inspections.javaee.Bundle;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class EjbJarUnusedSecurityRoleInspection extends AbstractEjbJarInspection {

  @NotNull
  public String getDisplayName() {
    return Bundle.message("inspection.name.KW_EjbJarUnusedSecurityRoleInspection");
  }

  @NotNull
  public String getShortName() {
    return "KW_EjbJarUnusedSecurityRoleInspection";
  }

  protected ProblemDescriptor[] checkDomFile(@NotNull DomFileElement<EjbJar> domFileElement, @NotNull InspectionManager manager, boolean isOnTheFly) {
    final List<ProblemDescriptor> problems = new ArrayList<ProblemDescriptor>();

    // collect role-names of security-roles
    final List<String> roleNames = new ArrayList<String>();
    for (SecurityRole securityRole : domFileElement.getRootElement().getAssemblyDescriptor().getSecurityRoles()) {
      roleNames.add(securityRole.getRoleName().getStringValue());
    }

    // throw away those that are used in a method-permission
    for (MethodPermission methodPermission : domFileElement.getRootElement().getAssemblyDescriptor().getMethodPermissions()) {
      final Boolean isUnchecked = methodPermission.getUnchecked().getValue();
      if (isUnchecked != null && isUnchecked) {
        return new ProblemDescriptor[0];
      }
      for (GenericDomValue<SecurityRole> name : methodPermission.getRoleNames()) {
        final SecurityRole role = name.getValue();
        if (role != null) {
          roleNames.remove(role.getRoleName().getStringValue());
        }
      }
    }

    // cache the role-names linked to from within a session or entity element.
    final Set<String> roleNamesReferredTo = new HashSet<String>();
    for (SessionBean sessionBean : domFileElement.getRootElement().getEnterpriseBeans().getSessions()) {
      for (SecurityRoleRef securityRoleRef : sessionBean.getSecurityRoleRefs()) {
        roleNamesReferredTo.add(securityRoleRef.getRoleLink().getValue());
      }
    }
    for (EntityBean entityBean : domFileElement.getRootElement().getEnterpriseBeans().getEntities()) {
      for (SecurityRoleRef securityRoleRef : entityBean.getSecurityRoleRefs()) {
        roleNamesReferredTo.add(securityRoleRef.getRoleLink().getValue());
      }
    }

    // create ProblemDescriptors
    for (SecurityRole securityRole : domFileElement.getRootElement().getAssemblyDescriptor().getSecurityRoles()) {
      final String roleName = securityRole.getRoleName().getValue();
      if (roleNames.contains((roleName))) {
        String message;
        if (roleNamesReferredTo.contains(roleName)) {
          message = Bundle.message("unused.but.referred.role-name", roleName);
        } else {
          message = Bundle.message("unused.role-name", roleName);
        }

        final XmlElement xmlElement = securityRole.getXmlElement();
        if (xmlElement != null) {
          problems.add(manager.createProblemDescriptor(xmlElement, message, REMOVE_QUICK_FIX, ProblemHighlightType.LIKE_UNUSED_SYMBOL));
        }
      }
    }

    return problems.toArray(new ProblemDescriptor[problems.size()]);
  }

  private LocalQuickFix REMOVE_QUICK_FIX = new RemoveQuickFix();

  private static class RemoveQuickFix implements LocalQuickFix {
    @NotNull
    public String getName() {
      return Bundle.message("quickfix.remove.undefined.security-role-reference");
    }

    public void applyFix(@NotNull Project project, ProblemDescriptor descriptor) {
      try {
        descriptor.getPsiElement().delete();
      }
      catch (IncorrectOperationException e) {
        // no actions
      }
    }

    @NotNull
    public String getFamilyName() {
      return getName();
    }
  }
}
