package com.konaworks.idea.inspections.javaee.web;

import com.intellij.codeInspection.InspectionManager;
import com.intellij.codeInspection.LocalQuickFix;
import com.intellij.codeInspection.ProblemDescriptor;
import com.intellij.codeInspection.ProblemHighlightType;
import com.intellij.javaee.model.xml.web.FilterMapping;
import com.intellij.javaee.model.xml.web.SecurityConstraint;
import com.intellij.javaee.model.xml.web.ServletMapping;
import com.intellij.javaee.model.xml.web.WebApp;
import com.intellij.javaee.model.xml.web.WebResourceCollection;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.XmlElement;
import com.intellij.psi.xml.XmlText;
import com.intellij.util.xml.DomFileElement;
import com.intellij.util.xml.GenericDomValue;
import com.konaworks.idea.inspections.javaee.Bundle;
import org.jetbrains.annotations.Nls;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.List;

public class UrlPatternInspection extends AbstractWebAppInspection {

  @Nls
  @NotNull
  public String getDisplayName() {
    return Bundle.message("inspection.name.KW_WebUrlPatternInspection");
  }

  @NonNls
  @NotNull
  public String getShortName() {
    return "KW_WebUrlPatternInspection";
  }


  protected ProblemDescriptor[] checkDomFile(@NotNull DomFileElement<WebApp> domFileElement, @NotNull InspectionManager manager, boolean isOnTheFly) {
    /*
    [copied from web.xml dtd/schema and servlet api spec :
    url-pattern must follow the rules specified in Section 11.2 of the Servlet API Specification.
    This pattern is assumed to be in URL-decoded form and must not contain CR(#xD) or LF(#xA).
    If it contains those characters, the container must inform the developer with a descriptive error message.
    */
    final List<ProblemDescriptor> problems = new ArrayList<ProblemDescriptor>();

    final List<ServletMapping> servletMappings = domFileElement.getRootElement().getServletMappings();
    for (ServletMapping servletMapping : servletMappings) {
      checkUrlPatterns(servletMapping.getUrlPatterns(), problems, manager);
    }

    final List<FilterMapping> list = domFileElement.getRootElement().getFilterMappings();
    for (FilterMapping filterMapping : list) {
      checkUrlPatterns(filterMapping.getUrlPatterns(), problems, manager);
    }

    final List<SecurityConstraint> constraints = domFileElement.getRootElement().getSecurityConstraints();
    for (SecurityConstraint constraint : constraints) {
      for (WebResourceCollection webResourceCollection : constraint.getWebResourceCollections()) {
        checkUrlPatterns(webResourceCollection.getUrlPatterns(), problems, manager);
      }
    }
    return problems.toArray(new ProblemDescriptor[problems.size()]);
  }

  private static void checkUrlPatterns(List<GenericDomValue<String>> list, List<ProblemDescriptor> problems, InspectionManager manager) {
    // todo ? quickfixes
    for (GenericDomValue<String> urlPatternValue : list) {
      final XmlElement urlPatternXmlElement = urlPatternValue.getXmlElement();
      if (urlPatternXmlElement != null) {
        final XmlText urlPatternXmlText = PsiTreeUtil.getChildOfAnyType(urlPatternXmlElement, XmlText.class);
        if (urlPatternXmlText != null) {
          final String urlPattern = urlPatternXmlText.getValue();
          if (urlPattern != null) {
            if ((urlPattern.indexOf('\r') != -1 || urlPattern.indexOf('\n') != -1)) {
              problems.add(manager.createProblemDescriptor(urlPatternXmlText, Bundle.message("invalid.url-pattern.CR_LF"),
                  (LocalQuickFix) null, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
            } else if (Character.isWhitespace(urlPattern.charAt(0)) || Character.isWhitespace(urlPattern.charAt(urlPattern.length() - 1))) {
              // Whitespace is of other/warning level, so might be moved to other inspection. but doesn't seem 
              // useful to have a separate inspection for that one, so we'll leave it this way.
              problems.add(manager.createProblemDescriptor(urlPatternXmlText, Bundle.message("invalid-url-pattern.whitespace"),
                  (LocalQuickFix) null, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
            }
          }
        }
      }
    }
  }
}
