package com.konaworks.idea.inspections.javaee.web;

import com.intellij.codeInspection.InspectionManager;
import com.intellij.codeInspection.LocalQuickFix;
import com.intellij.codeInspection.ProblemDescriptor;
import com.intellij.codeInspection.ProblemHighlightType;
import com.intellij.javaee.model.xml.SecurityRole;
import com.intellij.javaee.model.xml.SecurityRoleRef;
import com.intellij.javaee.model.xml.web.AuthConstraint;
import com.intellij.javaee.model.xml.web.SecurityConstraint;
import com.intellij.javaee.model.xml.web.Servlet;
import com.intellij.javaee.model.xml.web.WebApp;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.XmlElement;
import com.intellij.psi.xml.XmlFile;
import com.intellij.psi.xml.XmlTag;
import com.intellij.psi.xml.XmlText;
import com.intellij.util.IncorrectOperationException;
import com.intellij.util.xml.DomFileElement;
import com.intellij.util.xml.DomManager;
import com.intellij.util.xml.GenericDomValue;
import com.konaworks.idea.inspections.javaee.Bundle;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.List;

public class WebAppSecurityRoleReferenceInspection extends AbstractWebAppInspection {
  @NotNull
  public String getDisplayName() {
    return Bundle.message("inspection.name.KW_WebAppSecurityRoleReferenceInspection");
  }

  @NotNull
  public String getShortName() {
    return "KW_WebAppSecurityRoleReferenceInspection";
  }

  protected ProblemDescriptor[] checkDomFile(@NotNull DomFileElement<WebApp> domFileElement, @NotNull InspectionManager manager, boolean isOnTheFly) {
    final List<ProblemDescriptor> problems = new ArrayList<ProblemDescriptor>();

    final List<String> roleNames = new ArrayList<String>();
    for (SecurityRole securityRole : domFileElement.getRootElement().getSecurityRoles()) {
      roleNames.add(securityRole.getRoleName().getValue());
    }

    final String version = getWebAppVersion(domFileElement.getRootElement());

    // 1 check role-name usage in servlet/security-role-ref
    for (Servlet servlet : domFileElement.getRootElement().getServlets()) {
      for (SecurityRoleRef securityRoleRef : servlet.getSecurityRoleRefs()) {
        if (securityRoleRef.getRoleLink() != null && !roleNames.contains(securityRoleRef.getRoleLink().getValue())) {
          final XmlElement xmlElement = securityRoleRef.getRoleLink().getXmlElement();

          if (xmlElement != null) {
            final PsiElement xmlText = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
            if (xmlText != null) {
              problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("undefined.role-name"), ADD_QUICK_FIX, ProblemHighlightType.LIKE_UNKNOWN_SYMBOL));
              // the remove quick fix is not supported here, because the use of it would always yield in an incorrect security-role-ref.
            }
          }
        }
      }
      if (servlet.getRunAs() != null) {
        final String roleName = servlet.getRunAs().getRoleName().getValue();
        if (roleName != null && !roleNames.contains(roleName)) {
          final XmlElement xmlElement = servlet.getRunAs().getRoleName().getXmlElement();
          if (xmlElement != null) {
            final PsiElement xmlText = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
            if (xmlText != null) {
              problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("undefined.role-name"), ADD_QUICK_FIX, ProblemHighlightType.LIKE_UNKNOWN_SYMBOL));
            }
          }
        }
      }
    }

    // 2 check role-name usage from security-constraint/auth-constraint
    for (SecurityConstraint securityConstraint : domFileElement.getRootElement().getSecurityConstraints()) {
      for (GenericDomValue<String> name : securityConstraint.getAuthConstraint().getRoleNames()) {
        if (!roleNames.contains(name.getValue())) {
          final XmlElement xmlElement = name.getXmlElement();
          assert xmlElement != null;
          final PsiElement textElem = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
          if ("*".equals(name.getValue())) {
            // the allowed usage of * depends on the version...
            if (version != null && Float.valueOf(version) < 2.3) {
              if (textElem != null) {
                problems.add(manager.createProblemDescriptor(textElem,
                  Bundle.message("specially.reserved.role-name.not.allowed.in.this.webApp.version"),
                  REMOVE_PARENT_QUICK_FIX, ProblemHighlightType.GENERIC_ERROR_OR_WARNING
                ));/* todo enhancement add "upgrade web-app version" QuickFix ?*/
              }
            } else if (securityConstraint.getAuthConstraint().getRoleNames().size() > 1) {
              // todo ? highlighting might be a bit too coarse
              problems.add(manager.createProblemDescriptor(
                  xmlElement.getParent()/*auth-constraint*/, Bundle.message("unneeded.specific.role-names"),
                  REPLACE_WITH_ALL_QUICKFIX, ProblemHighlightType.GENERIC_ERROR_OR_WARNING
              ));
            }
          } else {
            if (textElem != null) {
              problems.add(manager.createProblemDescriptor(
                textElem, Bundle.message("undefined.role-name"),
                  new LocalQuickFix[]{ADD_QUICK_FIX, REMOVE_PARENT_QUICK_FIX}, ProblemHighlightType.LIKE_UNKNOWN_SYMBOL));
            }
          }
        }
      }
    }

    return problems.toArray(new ProblemDescriptor[problems.size()]);
  }

  private static final LocalQuickFix ADD_QUICK_FIX = new AddQuickFix();

  private static class AddQuickFix implements LocalQuickFix {
    @NotNull
    public String getName() {
      return Bundle.message("quickfix.add.undefined.security-role");
    }

    public void applyFix(@NotNull Project project, ProblemDescriptor descriptor) {
      final PsiFile file = descriptor.getPsiElement().getContainingFile();
      final DomFileElement<WebApp> domFileElement = DomManager.getDomManager(file.getProject()).getFileElement((XmlFile) file, WebApp.class);
      assert domFileElement != null;
      final WebApp webApp = domFileElement.getRootElement();

      webApp.addSecurityRole().getRoleName().setValue((/*(XmlText)*/ descriptor.getPsiElement()).getText());
    }

    @NotNull
    public String getFamilyName() {
      return getName();
    }
  }

  private static final LocalQuickFix REMOVE_PARENT_QUICK_FIX = new RemoveQuickFix();

  private static class RemoveQuickFix implements LocalQuickFix {
    @NotNull
    public String getName() {
      return Bundle.message("quickfix.remove.undefined.security-role-reference");
    }

    public void applyFix(@NotNull Project project, ProblemDescriptor descriptor) {
      try {
        descriptor.getPsiElement().getParent().delete();
      }
      catch (IncorrectOperationException e) {
        // no action
      }
    }

    @NotNull
    public String getFamilyName() {
      return getName();
    }
  }

  private static final LocalQuickFix REPLACE_WITH_ALL_QUICKFIX = new ReplaceWithAllQuickFix();

  private static class ReplaceWithAllQuickFix implements LocalQuickFix {

    @NotNull
    public String getName() {
      return Bundle.message("quickfix.replace.role-name.elements.with.star");
    }

    @NotNull
    public String getFamilyName() {
      return getName();
    }

    public void applyFix(@NotNull Project project, ProblemDescriptor descriptor) {
      // descriptor.getPsiElement is the auth-constraint
      final XmlTag authConstraintXmlElement = (XmlTag) descriptor.getPsiElement();
      final AuthConstraint authConstraint = (AuthConstraint) DomManager.getDomManager(project).getDomElement(authConstraintXmlElement);
      assert authConstraint != null;

      // remove all role-name children, except the '*' item
      final List<GenericDomValue<String>> roleNames = authConstraint.getRoleNames();
      try {
        for (GenericDomValue<String> roleName : roleNames) {
          if (!"*".equals(roleName.getValue())) {
            final XmlElement element = roleName.getXmlElement();
            if (element != null) {
              element.delete();
            }
          }
        }
      } catch (IncorrectOperationException e) {
        // no actions
      }
    }
  }
}
