package com.konaworks.idea.inspections.javaee.web;

import com.intellij.codeHighlighting.HighlightDisplayLevel;
import com.intellij.codeInspection.InspectionManager;
import com.intellij.codeInspection.LocalQuickFix;
import com.intellij.codeInspection.ProblemDescriptor;
import com.intellij.codeInspection.ProblemHighlightType;
import com.intellij.javaee.model.xml.SecurityRole;
import com.intellij.javaee.model.xml.SecurityRoleRef;
import com.intellij.javaee.model.xml.web.SecurityConstraint;
import com.intellij.javaee.model.xml.web.Servlet;
import com.intellij.javaee.model.xml.web.WebApp;
import com.intellij.openapi.project.Project;
import com.intellij.psi.xml.XmlElement;
import com.intellij.util.IncorrectOperationException;
import com.intellij.util.xml.DomFileElement;
import com.intellij.util.xml.GenericDomValue;
import com.konaworks.idea.inspections.javaee.Bundle;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class WebAppUnusedSecurityRoleInspection extends AbstractWebAppInspection {
  @NotNull
  public String getDisplayName() {
    return Bundle.message("inspection.name.KW_WebAppUnusedSecurityRoleInspection");
  }

  @NotNull
  public String getShortName() {
    return "KW_WebAppUnusedSecurityRoleInspection";
  }


  @NotNull
  @Override
  public HighlightDisplayLevel getDefaultLevel() {
    return HighlightDisplayLevel.WARNING;
  }

  protected ProblemDescriptor[] checkDomFile(@NotNull DomFileElement<WebApp> domFileElement, @NotNull InspectionManager manager, boolean isOnTheFly) {
    final List<ProblemDescriptor> problems = new ArrayList<ProblemDescriptor>();

    // collect role-names of security-roles
    final List<String> roleNames = new ArrayList<String>();
    for (SecurityRole securityRole : domFileElement.getRootElement().getSecurityRoles()) {
      roleNames.add(securityRole.getRoleName().getValue());
    }

    // throw away those that are used in a security-constraint/auth-constraint
    for (SecurityConstraint securityConstraint : domFileElement.getRootElement().getSecurityConstraints()) {
      for (GenericDomValue<String> name : securityConstraint.getAuthConstraint().getRoleNames()) {
        if ("*".equals(name.getValue())) {
          return new ProblemDescriptor[0];
        }
        if (roleNames.contains(name.getValue())) {
          roleNames.remove(name.getValue());
        }
      }
    }

    if (roleNames.size() == 0) {
      return new ProblemDescriptor[0];
    }

    // cache the role-names linked to from a servlet/security-role-ref/role-link
    // throw away those that are used in a or servlet/run-as
    final Set<String> roleNamesReferredTo = new HashSet<String>();
    for (Servlet servlet : domFileElement.getRootElement().getServlets()) {
      for (SecurityRoleRef securityRoleRef : servlet.getSecurityRoleRefs()) {
        roleNamesReferredTo.add(securityRoleRef.getRoleLink().getValue());
      }
      if (servlet.getRunAs() != null) {
        final String roleName = servlet.getRunAs().getRoleName().getValue();
        if (roleName != null) {
          roleNames.remove(roleName);
        }
      }
    }

    if (roleNames.size() == 0) {
      return new ProblemDescriptor[0];
    }

    // create ProblemDescriptors
    for (SecurityRole securityRole : domFileElement.getRootElement().getSecurityRoles()) {
      final String roleName = securityRole.getRoleName().getValue();
      if (roleNames.contains(roleName)) {
        String message;
        if (roleNamesReferredTo.contains(roleName)) {
          message = Bundle.message("unused.but.referred.role-name", roleName);
        } else {
          message = Bundle.message("unused.role-name", roleName);
        }
        final XmlElement securityRoleXmlElement = securityRole.getXmlElement();
        if (securityRoleXmlElement != null) {
          problems.add(manager.createProblemDescriptor(securityRoleXmlElement, message, REMOVE_QUICK_FIX, ProblemHighlightType.LIKE_UNUSED_SYMBOL));
        }
      }
    }

    return problems.toArray(new ProblemDescriptor[problems.size()]);
  }

  private static final LocalQuickFix REMOVE_QUICK_FIX = new RemoveQuickFix();

  private static class RemoveQuickFix implements LocalQuickFix {
    @NotNull
    public String getName() {
      return Bundle.message("quickfix.remove.undefined.security-role-reference");
    }

    public void applyFix(@NotNull Project project, ProblemDescriptor descriptor) {
      try {
        descriptor.getPsiElement().delete();
      }
      catch (IncorrectOperationException e) {
        // no actions
      }
    }

    @NotNull
    public String getFamilyName() {
      return getName();
    }
  }

}
