package com.konaworks.idea.inspections.javaee.web;

import com.intellij.codeHighlighting.HighlightDisplayLevel;
import com.intellij.codeInspection.InspectionManager;
import com.intellij.codeInspection.LocalQuickFix;
import com.intellij.codeInspection.ProblemDescriptor;
import com.intellij.codeInspection.ProblemHighlightType;
import com.intellij.javaee.model.xml.DisplayName;
import com.intellij.javaee.model.xml.Listener;
import com.intellij.javaee.model.xml.web.Filter;
import com.intellij.javaee.model.xml.web.SecurityConstraint;
import com.intellij.javaee.model.xml.web.Servlet;
import com.intellij.javaee.model.xml.web.WebApp;
import com.intellij.javaee.model.xml.web.WebResourceCollection;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiElement;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.XmlElement;
import com.intellij.psi.xml.XmlText;
import com.intellij.util.IncorrectOperationException;
import com.intellij.util.xml.DomFileElement;
import com.intellij.util.xml.GenericDomValue;
import com.konaworks.idea.inspections.javaee.Bundle;
import com.konaworks.idea.inspections.javaee.DuplicateNamedElementsProcessor;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.util.ArrayList;
import java.util.List;

public class WebAppWarningDuplicatesInspection extends AbstractWebAppInspection {
  private JPanel panel;

  private JCheckBox filterDisplayNameCheckBox;
  private JCheckBox listenerDisplayNameCheckBox;
  private JCheckBox servletDisplayNameCheckBox;
  private JCheckBox securityConstraintDisplayNameCheckBox;
  private JCheckBox messageDestinationDisplayNameCheckBox;

  private JCheckBox securityConstraintUrlPatternCheckBox;
  private JCheckBox securityConstraintRoleNameCheckBox;

  // settings readible by DefaultJDOMExternalizer
  public boolean checkFilterDisplayName = true;
  public boolean checkListenerDisplayName = true;
  public boolean checkServletDisplayName = true;
  public boolean checkSecurityConstraintDisplayName = true;
  public boolean checkMessageDestinationDisplayName = true;

  public boolean checkSecurityConstraintUrlPattern = true;
  public boolean checkSecurityConstraintRoleName = true;

  public WebAppWarningDuplicatesInspection() {
    filterDisplayNameCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkFilterDisplayName = filterDisplayNameCheckBox.isSelected();
      }
    });
    listenerDisplayNameCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkListenerDisplayName = listenerDisplayNameCheckBox.isSelected();
      }
    });
    servletDisplayNameCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkServletDisplayName = servletDisplayNameCheckBox.isSelected();
      }
    });
    securityConstraintDisplayNameCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkSecurityConstraintDisplayName = securityConstraintDisplayNameCheckBox.isSelected();
      }
    });
    messageDestinationDisplayNameCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkMessageDestinationDisplayName = messageDestinationDisplayNameCheckBox.isSelected();
      }
    });

    securityConstraintUrlPatternCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkSecurityConstraintUrlPattern = securityConstraintUrlPatternCheckBox.isSelected();
      }
    });
    securityConstraintRoleNameCheckBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        checkSecurityConstraintRoleName = securityConstraintRoleNameCheckBox.isSelected();
      }
    });
  }

  @NotNull
  public String getDisplayName() {
    return Bundle.message("inspection.name.KW_WebAppWarningDuplicatesInspection");
  }

  @NotNull
  public String getShortName() {
    return "KW_WebAppWarningDuplicatesInspection";
  }

  @Nullable
  public JComponent createOptionsPanel() {
    filterDisplayNameCheckBox.setSelected(checkFilterDisplayName);
    listenerDisplayNameCheckBox.setSelected(checkListenerDisplayName);
    servletDisplayNameCheckBox.setSelected(checkServletDisplayName);
    securityConstraintDisplayNameCheckBox.setSelected(checkSecurityConstraintDisplayName);
    messageDestinationDisplayNameCheckBox.setSelected(checkMessageDestinationDisplayName);
    securityConstraintUrlPatternCheckBox.setSelected(checkSecurityConstraintUrlPattern);
    securityConstraintRoleNameCheckBox.setSelected(checkSecurityConstraintRoleName);

    return panel;
  }

  @NotNull
  @Override
  public HighlightDisplayLevel getDefaultLevel() {
    return HighlightDisplayLevel.INFO;
  }

  protected ProblemDescriptor[] checkDomFile(@NotNull DomFileElement<WebApp> domFileElement, @NotNull InspectionManager manager, boolean isOnTheFly) {
    final List<ProblemDescriptor> problems = new ArrayList<ProblemDescriptor>();
    // todo: (minor) BUG duplicate display-name elements may be used in combination with different xml:lang attributes. This inspection doesn't treat those as different... (but not expected to be used much)
    // remark: EAP 5581 has a bug, because it marks the xml:lang attribute as invalid there... todo report bug in JB Jira.
    // todo NOT NOW optional extra inspection: multiple display-name elements within one declaration withou xml:lang attributes... (xml:lang is not expected to be used much, so not now)
    if (checkFilterDisplayName) {
      final DuplicateNamedElementsProcessor<DisplayName> processor = new DuplicateNamedElementsProcessor<DisplayName>();
      for (Filter filter : domFileElement.getRootElement().getFilters()) {
        for (DisplayName displayName : filter.getDisplayNames()) {
          processor.put(displayName.getValue(), displayName);
        }
      }
      for (DisplayName displayName : processor.getDuplicates()) {
        final XmlElement xmlElement = displayName.getXmlElement();

        if (xmlElement != null) {
          final PsiElement xmlText = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
          if (xmlText != null) {
            problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("duplicate.filter.display-name"), (LocalQuickFix) null, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
          }
        }
      }
    }
    if (checkListenerDisplayName) {
      final DuplicateNamedElementsProcessor<DisplayName> processor = new DuplicateNamedElementsProcessor<DisplayName>();
      for (Listener listener : domFileElement.getRootElement().getListeners()) {
        for (DisplayName displayName : listener.getDisplayNames()) {
          processor.put(displayName.getValue(), displayName);
        }
      }
      for (DisplayName displayName : processor.getDuplicates()) {
        final XmlElement xmlElement = displayName.getXmlElement();
        if (xmlElement != null) {
          final PsiElement xmlText = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
          if (xmlText != null) {
            problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("duplicate.listener.display-name"), (LocalQuickFix) null, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
          }
        }
      }
    }
    if (checkServletDisplayName) {
      final DuplicateNamedElementsProcessor<DisplayName> processor = new DuplicateNamedElementsProcessor<DisplayName>();
      for (Servlet servlet : domFileElement.getRootElement().getServlets()) {
        for (DisplayName displayName : servlet.getDisplayNames()) {
          processor.put(displayName.getValue(), displayName);
        }
      }
      for (DisplayName displayName : processor.getDuplicates()) {
        final XmlElement xmlElement = displayName.getXmlElement();
        if (xmlElement != null) {
          final PsiElement xmlText = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
          if (xmlText != null) {
            problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("duplicate.servlet.display-name"), (LocalQuickFix) null, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
          }
        }
      }
    }
    if (checkSecurityConstraintDisplayName) {
      final DuplicateNamedElementsProcessor<DisplayName> processor = new DuplicateNamedElementsProcessor<DisplayName>();
      for (SecurityConstraint constraint : domFileElement.getRootElement().getSecurityConstraints()) {
        for (DisplayName displayName : constraint.getDisplayNames()) {
          processor.put(displayName.getValue(), displayName);
        }
      }
      for (DisplayName displayName : processor.getDuplicates()) {
        final XmlElement xmlElement = displayName.getXmlElement();
        if (xmlElement != null) {
          final PsiElement textElem = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
          if (textElem != null) {
            problems.add(manager.createProblemDescriptor(textElem, Bundle.message("duplicate.security-constraint.display-name"), (LocalQuickFix) null, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
          }
        }
      }
    }

    if (checkSecurityConstraintUrlPattern || checkSecurityConstraintRoleName) {
      for (SecurityConstraint securityConstraint : domFileElement.getRootElement().getSecurityConstraints()) {
        if (checkSecurityConstraintUrlPattern) {
          for (WebResourceCollection webResourceCollection : securityConstraint.getWebResourceCollections()) {
            final DuplicateNamedElementsProcessor<GenericDomValue<String>> processor = new DuplicateNamedElementsProcessor<GenericDomValue<String>>();
            for (GenericDomValue<String> genericDomValue : webResourceCollection.getUrlPatterns()) {
              processor.put(genericDomValue.getValue(), genericDomValue);
            }
            for (GenericDomValue<String> duplicate : processor.getDuplicates()) {
              final XmlElement xmlElement = duplicate.getXmlElement();
              if (xmlElement != null) {
                final PsiElement xmlText = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
                if (xmlText != null) {
                  problems.add(manager.createProblemDescriptor(xmlText, Bundle.message("duplicate.url-pattern"), DELETE_PARENT_FIX, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
                }
              }
            }
          }
        }
        if (checkSecurityConstraintRoleName) {
          final DuplicateNamedElementsProcessor<GenericDomValue<String>> processor = new DuplicateNamedElementsProcessor<GenericDomValue<String>>();
          for (GenericDomValue<String> roleName : securityConstraint.getAuthConstraint().getRoleNames()) {
            processor.put(roleName.getValue(), roleName);
          }
          for (GenericDomValue<String> duplRoleName : processor.getDuplicates()) {
            final XmlElement xmlElement = duplRoleName.getXmlElement();
            if (xmlElement != null) {
              final PsiElement textElem = PsiTreeUtil.getChildOfAnyType(xmlElement, XmlText.class);
              if (textElem != null) {
                problems.add(manager.createProblemDescriptor(textElem, Bundle.message("duplicate.role-name"), DELETE_PARENT_FIX, ProblemHighlightType.GENERIC_ERROR_OR_WARNING));
              }
            }
          }
        }
      }
    }

    return problems.toArray(new ProblemDescriptor[problems.size()]);
  }

  private static final LocalQuickFix DELETE_PARENT_FIX = new MyQuickFix();

  private static class MyQuickFix implements LocalQuickFix {
    @NotNull
    public String getName() {
      return Bundle.message("quickfix.remove.duplicate.url-pattern");
    }

    public void applyFix(@NotNull Project project, ProblemDescriptor descriptor) {
      try {
        descriptor.getPsiElement().delete();
      }
      catch (IncorrectOperationException e) {
        // no actions
      }
    }

    @NotNull
    public String getFamilyName() {
      return getName();
    }
  }
}
